from typing import List
from asf_search.exceptions import ASFAuthenticationError, ASFSearch4xxError, ASFSearch5xxError

from ASFProduct.test_ASFProduct import run_test_ASFProduct_Geo_Search, run_test_stack
from ASFSession.test_ASFSession import run_auth_with_cookiejar, run_auth_with_creds, run_auth_with_token, run_test_asf_session_rebuild_auth
from BaselineSearch.test_baseline_search import *
from Search.test_search import run_test_ASFSearchResults, run_test_search, run_test_search_http_error
from CMR.test_MissionList import run_test_get_project_names


from pytest import raises
from unittest.mock import patch

import os
import pathlib
import yaml

import requests
from tests.BaselineSearch.Stack.test_stack import run_test_find_new_reference, run_test_get_baseline_from_stack, run_test_get_default_product_type, run_test_valid_state_vectors

from tests.download.test_download import run_test_download_url_auth_error

# asf_search.ASFProduct Tests
def test_ASFProduct(**args) -> None:
    """
    Tests Basic ASFProduct with mock searchAPI response
    """
    test_info = args["test_info"]
    geographic_response = get_resource(test_info["products"])
    run_test_ASFProduct_Geo_Search(geographic_response)

def test_ASFProduct_Stack(**args) -> None:
    """
    Tests ASFProduct.stack() with reference and corresponding stack
    Checks for temporalBaseline order, 
    asserting the stack is ordered by the scene's temporalBaseline (in ascending order)
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    preprocessed_stack = get_resource(test_info["preprocessed_stack"])
    processed_stack = get_resource(test_info["processed_stack"])
    run_test_stack(reference, preprocessed_stack, processed_stack)
    
# asf_search.ASFSession Tests
def test_ASFSession_Error(**args) -> None:
    """
    Test ASFSession.auth_with_creds for sign in errors
    """
    test_info = args["test_info"]
    username = test_info["username"]
    password = test_info["password"]
    with patch('asf_search.ASFSession.get') as mock_get:
        mock_get.return_value = "Error"

        with raises(ASFAuthenticationError):
            run_auth_with_creds(username, password)

def test_ASFSession_Token_Error(**args) -> None:
    """
    Test ASFSession.auth_with_token for sign in errors
    """
    test_info = args["test_info"]
    token = test_info["token"]

    with raises(ASFAuthenticationError):
        run_auth_with_token(token)

def test_ASFSession_Cookie_Error(**args) -> None:
    """
    Test ASFSession.auth_with_cookie for sign in errors
    """
    test_info = args["test_info"]
    cookies = test_info["cookies"]

    with raises(ASFAuthenticationError):
        run_auth_with_cookiejar(cookies)

def test_asf_session_rebuild_auth(**args) -> None:
    """
    Test asf_search.ASFSession.rebuild_auth
    When redirecting from an ASF domain, only accept 
    domains listed in ASFSession.AUTH_DOMAINS
    """
    test_info = args["test_info"]
    original_domain = test_info["original_domain"]
    response_domain = test_info["response_domain"]
    response_code = test_info["response_code"]
    final_token = test_info["final_token"]

    run_test_asf_session_rebuild_auth(original_domain, response_domain, response_code, final_token)

# asf_search.search.baseline_search Tests
def test_get_preprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_params with a reference scene
    that's part of a pre-calculated platform, asserting that get_stack_params returns an object with two parameters
    \n1. processingLevel
    \n2. insarStackId
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])

    run_test_get_preprocessed_stack_params(reference)

def test_get_unprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_params with a reference scene
    that's not part of a pre-calculated platform, asserting that get_stack_params returns an object with seven parameters
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])

    run_test_get_unprocessed_stack_params(reference)

def test_get_stack_params_invalid_insarStackId(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_params with a the reference scene's 
    insarStackID set to an invalid value, and asserting an ASFBaselineError is raised
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    
    run_get_stack_params_invalid_insarStackId(reference)
    
def test_get_stack_params_invalid_platform(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_params with a the reference scene's 
    platform set to an invalid value, and asserting an ASFBaselineError is raised
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])    
    run_test_get_stack_params_invalid_platform_raises_error(reference)
    
def test_temporal_baseline(**args) -> None:
    """
    Test asf_search.search.baseline_search.calc_temporal_baselines, asserting mutated baseline stack
    is still the same length and that each product's properties contain a temporalBaseline key 
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    stack = get_resource(test_info["stack"])
    run_test_calc_temporal_baselines(reference, stack)
    
def test_stack_from_product(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_product, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    stack = get_resource(test_info["stack"])
    
    run_test_stack_from_product(reference, stack)

def test_stack_from_id(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_id, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args["test_info"]
    stack_id = test_info["stack_id"]
    stack_reference_data = test_info["stack_reference"]
    stack_data = test_info["stack"]

    stack_reference = get_resource(stack_reference_data)
    stack = []

    if(stack_data != []):
        stack = get_resource(stack_data)

    run_test_stack_from_id(stack_id, stack_reference, stack)

# asf_search.ASFSearchResults Tests
def test_ASFSearchResults(**args) -> None:
    """
    Test asf_search.ASFSearchResults, asserting initialized values, 
    and geojson response returns object with type FeatureCollection
    """
    test_info = args["test_info"]
    search_response = get_resource(test_info["response"])

    run_test_ASFSearchResults(search_response)

# asf_search.search Tests
def test_ASFSearch_Search(**args) -> None:
    """
    Test asf_search.search, asserting returned value is expected result
    """
    test_info = args["test_info"]
    parameters = get_resource(test_info["parameters"])
    answer = get_resource(test_info["answer"])

    run_test_search(parameters, answer)
    
def test_ASFSearch_Search_Error(**args) -> None:
    """
    Test asf_search.search errors,
    asserting server and client errors are raised
    """
    test_info = args["test_info"]
    parameters = test_info["parameters"]
    report = test_info["report"]
    error_code = test_info["status_code"]
    if error_code == 400:
        with raises(ASFSearch4xxError):
            run_test_search_http_error(parameters, error_code, report)
    if error_code == 500:
        with raises(ASFSearch5xxError):
            run_test_search_http_error(parameters, error_code, report)

def test_get_platform_campaign_names(**args) -> None:
    test_info = args["test_info"]
    cmr_ummjson = get_resource(test_info["cmr_ummjson"])
    campaigns: List[str] = get_resource(test_info["campaigns"])
    
    run_test_get_project_names(cmr_ummjson, campaigns)

def test_download_url(**args) -> None:
    """
    Test asf_search.download.download_url
    """
    test_info = args["test_info"]
    url = test_info["url"]
    path = test_info["path"]
    filename = test_info["filename"]
        
    if filename == "error":
        run_test_download_url_auth_error(url, path, filename)

def test_find_new_reference(**args) -> None:
    """
    Test asf_search.baseline.calc.find_new_reference
    """
    test_info = args["test_info"]
    stack = get_resource(test_info["stack"])
    output_index = get_resource(test_info["output_index"])
    
    run_test_find_new_reference(stack, output_index)

def test_get_default_product_type(**args) -> None:
    test_info = args["test_info"]
    scene_name = get_resource(test_info["scene_name"])
    product_type = get_resource(test_info["product_type"])
    
    run_test_get_default_product_type(scene_name, product_type)

def test_get_baseline_from_stack(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['reference'])
    stack = get_resource(test_info['stack'])
    output_stack = get_resource(test_info['output_stack'])
    error = get_resource(test_info['error'])
    run_test_get_baseline_from_stack(reference, stack, output_stack, error)

def test_valid_state_vectors(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['reference'])
    output = get_resource(test_info['output'])
    
    run_test_valid_state_vectors(reference, output)
# Testing resource loading utilities

# Finds and loads file from yml_tests/Resouces/ if loaded field ends with .yml/yaml extension
def get_resource(yml_file):
    
    if isinstance(yml_file, str):
        if yml_file.endswith((".yml", ".yaml")):
            base_path = pathlib.Path(__file__).parent.resolve()
            with open(os.path.join(base_path, "yml_tests", "Resources", yml_file), "r") as f:
                try:
                    return yaml.safe_load(f)
                except yaml.YAMLError as exc:
                    print(exc)
    elif isinstance(yml_file, List): #check if it's a list of yml files
        if len(yml_file) > 0:
            if isinstance(yml_file[0], str):
                if yml_file[0].endswith((".yml", ".yaml")):
                    return [get_resource(file) for file in yml_file]

    return yml_file



from math import sqrt, cos, sin, radians
from typing import List

import numpy as np
from dateutil.parser import parse

from asf_search import ASFProduct
# WGS84 constants
a = 6378137
f = pow((1.0 - 1 / 298.257224), 2)
# Technically f is normally considered to just be that 298... part but this is all we ever use, so
# pre-calc and cache and call it all f anyhow

def calculate_perpendicular_baselines(reference: str, stack: List[ASFProduct]):
    for product in stack:
        baselineProperties = product.baseline
        positionProperties = baselineProperties['stateVectors']['positions']

        if None in [positionProperties['prePositionTime'], positionProperties['postPositionTime'], positionProperties['prePosition'], positionProperties['postPosition']]:
            baselineProperties['noStateVectors'] = True
            continue

        asc_node_time = parse(baselineProperties['ascendingNodeTime']).timestamp()

        start = parse(product.properties['startTime']).timestamp()
        end = parse(product.properties['stopTime']).timestamp()
        center = start + ((end - start) / 2)
        baselineProperties['relative_start_time'] = start - asc_node_time
        baselineProperties['relative_center_time'] = center - asc_node_time
        baselineProperties['relative_end_time'] = end - asc_node_time

        t_pre = parse(positionProperties['prePositionTime']).timestamp()
        t_post = parse(positionProperties['postPositionTime']).timestamp()
        product.baseline['relative_sv_pre_time'] = t_pre - asc_node_time
        product.baseline['relative_sv_post_time'] = t_post - asc_node_time

    for product in stack:
        # product.properties['granulePosition'] = get_granule_position(reference.properties['centerLat'], reference.properties['centerLon'])
        
        if product.properties['sceneName'] == reference:
            reference = product
            reference.properties['perpendicularBaseline'] = 0
            # Cache these values
            reference.baseline['granulePosition'] = get_granule_position(reference.properties['centerLat'], reference.properties['centerLon'])
            break

    for secondary in stack:
        if secondary.baseline.get('noStateVectors'):
            secondary.properties['perpendicularBaseline'] = None
            continue

        shared_rel_time = get_shared_sv_time(reference, secondary)

        reference_shared_pos = get_pos_at_rel_time(reference, shared_rel_time)
        reference_shared_vel = get_vel_at_rel_time(reference, shared_rel_time)
        secondary_shared_pos = get_pos_at_rel_time(secondary, shared_rel_time)
        #secondary_shared_vel = get_vel_at_rel_time(secondary, shared_rel_time) # unused

        # need to get sat pos and sat vel at center time
        reference.baseline['alongBeamVector'] = get_along_beam_vector(reference_shared_pos, reference.baseline['granulePosition'])
        reference.baseline['upBeamVector'] = get_up_beam_vector(reference_shared_vel, reference.baseline['alongBeamVector'])

        perpendicular_baseline = get_paired_granule_baseline(
            reference.baseline['granulePosition'],
            reference.baseline['upBeamVector'],
            secondary_shared_pos)
        if abs(perpendicular_baseline) > 100000:
            perpendicular_baseline = None
        secondary.properties['perpendicularBaseline'] = perpendicular_baseline

    return stack

# Convert granule center lat/lon to fixed earth coordinates in meters using WGS84 ellipsoid.
def get_granule_position(scene_center_lat, scene_center_lon):
    lat = radians(float(scene_center_lat))
    lon = radians(float(scene_center_lon))
    coslat = cos(lat) # This value gets used a couple times, cache it
    sinlat = sin(lat) # This value gets used a couple times, cache it
    C = 1.0 / (sqrt(pow(coslat, 2) + f * pow(sinlat, 2)))
    S = f * C
    aC = a * C
    granule_position = np.array([aC * coslat * cos(lon), aC * coslat * sin(lon), a * S * sinlat])
    return(granule_position)

# Calculate along beam vector from sat pos and granule pos
def get_along_beam_vector(satellite_position, granule_position):
    along_beam_vector = np.subtract(satellite_position, granule_position)
    along_beam_vector = np.divide(along_beam_vector, np.linalg.norm(along_beam_vector)) # normalize
    return(along_beam_vector)

# Calculate up beam vector from sat velocity and along beam vector
def get_up_beam_vector(satellite_velocity, along_beam_vector):
    up_beam_vector = np.cross(satellite_velocity, along_beam_vector)
    up_beam_vector = np.divide(up_beam_vector, np.linalg.norm(up_beam_vector)) # normalize
    return(up_beam_vector)

# Calculate baseline between reference and paired granule
def get_paired_granule_baseline(reference_granule_position, reference_up_beam_vector, paired_satellite_position):
    posd = np.subtract(paired_satellite_position, reference_granule_position)
    baseline = np.dot(reference_up_beam_vector, posd)
    return(int(round(baseline)))

# Find a relative orbit time covered by both granules' SVs
def get_shared_sv_time(reference, secondary):
    start = max(reference.baseline['relative_sv_pre_time'], secondary.baseline['relative_sv_pre_time'])
    end = max(reference.baseline['relative_sv_post_time'], secondary.baseline['relative_sv_post_time'])

    # Favor the start/end SV time of the reference so we can use that SV directly without interpolation
    if start == reference.baseline['relative_sv_pre_time']:
        return start
    if end == reference.baseline['relative_sv_post_time']:
        return end

    return start

# Interpolate a position SV based on relative time
def get_pos_at_rel_time(granule: ASFProduct, relative_time):
    if relative_time == granule.baseline['relative_sv_pre_time']:
        return granule.baseline['stateVectors']['positions']['prePosition']
    if relative_time == granule.baseline['relative_sv_post_time']:
        return granule.baseline['stateVectors']['positions']['postPosition']

    duration = granule.baseline['relative_sv_post_time'] - granule.baseline['relative_sv_pre_time']
    factor = (relative_time - granule.baseline['relative_sv_pre_time']) / duration

    vec_a = granule.baseline['stateVectors']['positions']['prePosition']
    vec_b = granule.baseline['stateVectors']['positions']['postPosition']

    v = [
        interpolate(vec_a[0], vec_b[0], factor),
        interpolate(vec_a[1], vec_b[1], factor),
        interpolate(vec_a[2], vec_b[2], factor)]

    return radius_fix(granule, v, relative_time)

# Interpolate a velocity SV based on relative time
def get_vel_at_rel_time(granule: ASFProduct, relative_time):
    velocityProperties = granule.baseline['stateVectors']['velocities']
    if relative_time == granule.baseline['relative_sv_pre_time']:
        return velocityProperties['preVelocity']
    if relative_time == granule.baseline['relative_sv_post_time']:
        return velocityProperties['postVelocity']

    duration = granule.baseline['relative_sv_post_time'] - granule.baseline['relative_sv_pre_time']
    factor = (relative_time - granule.baseline['relative_sv_pre_time']) / duration

    vec_a = velocityProperties['preVelocity']
    vec_b = velocityProperties['postVelocity']

    v = [
        interpolate(vec_a[0], vec_b[0], factor),
        interpolate(vec_a[1], vec_b[1], factor),
        interpolate(vec_a[2], vec_b[2], factor)]

    return v

# convenience 1d linear interp
def interpolate(p0, p1, x):
    return (p0 * (1.0 - x)) + (p1 * x)

# Bump the provided sat pos out to a radius interpolated between the start and end sat pos vectors
def radius_fix(granule: ASFProduct, sat_pos, relative_time):
    positionProperties = granule.baseline['stateVectors']['positions']
    pre_l = np.linalg.norm(positionProperties['prePosition'])
    post_l = np.linalg.norm(positionProperties['postPosition'])
    sat_pos_l = np.linalg.norm(sat_pos)
    dt = relative_time - granule.baseline['relative_sv_pre_time']
    new_l = pre_l + (post_l - pre_l) * dt / (granule.baseline['relative_sv_post_time'] - granule.baseline['relative_sv_pre_time'])
    sat_pos[0] = sat_pos[0] * new_l / sat_pos_l
    sat_pos[1] = sat_pos[1] * new_l / sat_pos_l
    sat_pos[2] = sat_pos[2] * new_l / sat_pos_l
    return sat_pos



from shapely.geometry import shape, Point, Polygon, mapping
import json
from collections import UserList
import requests

from asf_search.download import download_url
from asf_search import ASFSession


class ASFProduct:
    def __init__(self, args: dict):
        self.properties = args['properties']
        self.geometry = args['geometry']
        self.baseline = args['baseline']

    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def geojson(self) -> dict:
        return {
            'type': 'Feature',
            'geometry': self.geometry,
            'properties': self.properties
        }

    def download(self, path: str, filename: str = None, session: ASFSession = None) -> None:
        """
        Downloads this product to the specified path and optional filename.

        :param path: The directory into which this product should be downloaded.
        :param filename: Optional filename to use instead of the original filename of this product.
        :param session: The session to use, in most cases should be authenticated beforehand

        :return: None
        """
        if filename is None:
            filename = self.properties['fileName']

        download_url(url=self.properties['url'], path=path, filename=filename, session=session)

    def stack(self) -> UserList:
        """
        Builds a baseline stack from this product.

        :return: ASFSearchResults(list) of the stack, with the addition of baseline values (temporal, perpendicular) attached to each ASFProduct.
        """
        from .search.baseline_search import stack_from_product

        return stack_from_product(self)

    def centroid(self) -> Point:
        """
        Finds the centroid of a product
        """
        coords = mapping(shape(self.geometry))['coordinates'][0]
        lons = [p[0] for p in coords]
        if max(lons) - min(lons) > 180:
            unwrapped_coords = [a if a[0] > 0 else [a[0] + 360, a[1]] for a in coords]
        else:
            unwrapped_coords = [a for a in coords]

        return Polygon(unwrapped_coords).centroid



from collections import UserList
from multiprocessing import Pool
import json
from asf_search import ASFSession


class ASFSearchResults(UserList):
    def geojson(self):
        return {
            'type': 'FeatureCollection',
            'features': [product.geojson() for product in self]
        }

    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def download(self, path: str, session: ASFSession = None, processes=1) -> None:
        """
        Iterates over each ASFProduct and downloads them to the specified path.

        :param path: The directory into which the products should be downloaded.
        :param session: The session to use, in most cases should be authenticated beforehand
        :param processes: Number of download processes to use. Defaults to 1 (i.e. sequential download)

        :return: None
        """
        if session is None:
            session = ASFSession()

        if processes == 1:
            for product in self:
                product.download(path=path, session=session)
        else:
            pool = Pool(processes=processes)
            args = [(product, path, session) for product in self]
            pool.map(_download_product, args)
            pool.close()
            pool.join()


def _download_product(args):
    product, path, session = args
    product.download(path=path, session=session)



from .search import search
from .granule_search import granule_search
from .product_search import product_search
from .geo_search import geo_search
from .baseline_search import stack_from_id
from .campaigns import campaigns



from asf_search.CMR.MissionList import get_campaigns
from asf_search.search.campaigns import _get_project_names
import pytest
import requests_mock

from asf_search.constants.INTERNAL import CMR_COLLECTIONS, CMR_HOST
from asf_search.exceptions import CMRError

            
def test_getMissions_error():
    with requests_mock.Mocker() as m:
        m.register_uri('POST', f"https://" + CMR_HOST + CMR_COLLECTIONS, status_code=300, json={'error': {'report': ""}})
        
        with pytest.raises(CMRError):    
            get_campaigns({})

def test_getMissions_error_parsing():
    with requests_mock.Mocker() as m:
        m.post(f"https://" + CMR_HOST + CMR_COLLECTIONS)
        
        with pytest.raises(CMRError):
            get_campaigns({})

def run_test_get_project_names(cmr_ummjson, campaigns):
    assert _get_project_names(cmr_ummjson) == campaigns



from numbers import Number
from asf_search.ASFProduct import ASFProduct
from asf_search.constants import INTERNAL
from asf_search.search import search

from asf_search.ASFSearchResults import ASFSearchResults

import requests_mock

def run_test_ASFSearchResults(search_resp):
    search_results = ASFSearchResults(map(ASFProduct, search_resp))

    assert(len(search_results) == len(search_resp))
    assert(search_results.geojson()['type'] == 'FeatureCollection')

    for (idx, feature) in enumerate(search_results.data):
        assert(feature.geojson()['geometry'] == search_resp[idx]['geometry'])
        assert(feature.geojson()['properties'] == search_resp[idx]['properties'])

def run_test_search(search_parameters, answer):
    with requests_mock.Mocker() as m:
        m.post(f"https://{search_parameters['host']}{INTERNAL.SEARCH_PATH}", json={'features': answer})
        response = search(**search_parameters)

        if search_parameters["maxResults"] != 'None':
            assert(len(response) == search_parameters["maxResults"])

        assert(len(response) == len(answer))
        # assert(response.geojson()["features"] == answer)

def run_test_search_http_error(search_parameters, status_code: Number, report: str):
    with requests_mock.Mocker() as m:
        m.register_uri('POST', f"https://{search_parameters['host']}{INTERNAL.SEARCH_PATH}", status_code=status_code, json={'error': {'report': report}})
        
        search(**search_parameters)



from asf_search.baseline.stack import get_baseline_from_stack

from asf_search.search import search
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFProduct import ASFProduct
from asf_search.search.product_search import product_search
from asf_search.constants import INTERNAL, PLATFORM
from asf_search.exceptions import ASFSearchError, ASFBaselineError


precalc_platforms = [
    PLATFORM.ALOS,
    PLATFORM.RADARSAT,
    PLATFORM.ERS1,
    PLATFORM.ERS2,
    PLATFORM.JERS]


def stack_from_product(
        reference: ASFProduct,
        strategy = None,
        host: str = INTERNAL.SEARCH_API_HOST,
        cmr_token: str = None,
        cmr_provider: str = None) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference ASFProduct

    :param reference: Reference scene to base the stack from, and from which to calculate perpendicular/temporal baselines
    :param strategy: If the requested reference can not be used to calculate perpendicular baselines, this sort function will be used to pick an alternative reference from the stack. 'None' implies that no attempt will be made to find an alternative reference.
    :param host: SearchAPI host, defaults to Production SearchAPI. This option is intended for dev/test purposes.
    :param cmr_token: EDL Auth Token for authenticated searches, see https://urs.earthdata.nasa.gov/user_tokens
    :param cmr_provider: Custom provider name to constrain CMR results to, for more info on how this is used, see https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#c-provider

    :return: ASFSearchResults(dict) of search results
    """

    stack_params = get_stack_params(reference)
    stack = search(**stack_params, host=host, cmr_token=cmr_token, cmr_provider=cmr_provider)
    stack, warnings = get_baseline_from_stack(reference=reference, stack=stack)

    stack.sort(key=lambda product: product.properties['temporalBaseline'])

    return stack

def stack_from_id(
        reference_id: str,
        strategy = None,
        host: str = INTERNAL.SEARCH_API_HOST,
        cmr_token: str = None,
        cmr_provider: str = None) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference product ID

    :param reference_id: Reference product to base the stack from, and from which to calculate perpendicular/temporal baselines
    :param strategy: If the requested reference can not be used to calculate perpendicular baselines, this sort function will be used to pick an alternative reference from the stack. 'None' implies that no attempt will be made to find an alternative reference.
    :param host: SearchAPI host, defaults to Production SearchAPI. This option is intended for dev/test purposes.
    :param cmr_token: EDL Auth Token for authenticated searches, see https://urs.earthdata.nasa.gov/user_tokens
    :param cmr_provider: Custom provider name to constrain CMR results to, for more info on how this is used, see https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#c-provider

    :return: ASFSearchResults(list) of search results
    """
    reference_results = product_search(
        [reference_id],
        host=host,
        cmr_token=cmr_token,
        cmr_provider=cmr_provider)

    if len(reference_results) <= 0:
        raise ASFSearchError(f'Reference product not found: {reference_id}')
    reference = reference_results[0]

    return stack_from_product(reference, host=host, cmr_token=cmr_token, cmr_provider=cmr_provider)


def get_stack_params(reference: ASFProduct) -> dict:

    stack_params = {
        'processingLevel': [reference.properties['processingLevel']]
    }

    if reference.properties['platform'] in precalc_platforms:
        if reference.properties['insarStackId'] not in [None, 'NA', 0, '0']:
            stack_params['insarStackId'] = reference.properties['insarStackId']
            return stack_params
        raise ASFBaselineError(f'Requested reference product needs a baseline stack ID but does not have one: {reference.properties["fileID"]}')

    # build a stack from scratch if it's a non-precalc dataset with state vectors
    if reference.properties['platform'] in [PLATFORM.SENTINEL1A, PLATFORM.SENTINEL1B]:
        stack_params['platform'] = [PLATFORM.SENTINEL1]
        stack_params['beamMode'] = [reference.properties['beamModeType']]
        stack_params['flightDirection'] = [reference.properties['flightDirection']]
        #stack_params['lookDirection'] = [ref_scene.properties['lookDirection']]
        stack_params['relativeOrbit'] = [int(reference.properties['pathNumber'])]  # path
        if reference.properties['polarization'] in ['HH', 'HH+HV']:
            stack_params['polarization'] = ['HH','HH+HV']
        elif reference.properties['polarization'] in ['VV', 'VV+VH']:
            stack_params['polarization'] = ['VV','VV+VH']
        else:
            stack_params['polarization'] = [reference.properties['polarization']]
        stack_params['intersectsWith'] = reference.centroid().wkt
        return stack_params

    raise ASFBaselineError(f'Reference product is not a pre-calculated baseline dataset, and not a known ephemeris-based dataset: {reference.properties["fileID"]}')



from typing import Union, Iterable, Tuple
import requests
from requests.exceptions import HTTPError
import datetime
import math

import warnings
import inspect

from asf_search import __version__
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFProduct import ASFProduct
from asf_search.exceptions import ASFSearch4xxError, ASFSearch5xxError, ASFServerError
from asf_search.constants import INTERNAL


def search(
        absoluteOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        asfFrame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        beamMode: Union[str, Iterable[str]] = None,
        collectionName: Union[str, Iterable[str]] = None,
        campaign: Union[str, Iterable[str]] = None,
        maxDoppler: float = None,
        minDoppler: float = None,
        end: Union[datetime.datetime, str] = None,
        maxFaradayRotation: float = None,
        minFaradayRotation: float = None,
        flightDirection: str = None,
        flightLine: str = None,
        frame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        granule_list: Union[str, Iterable[str]] = None,
        groupID: Union[str, Iterable[str]] = None,
        insarStackId: str = None,
        instrument: Union[str, Iterable[str]] = None,
        intersectsWith: str = None,
        lookDirection: Union[str, Iterable[str]] = None,
        offNadirAngle: Union[float, Tuple[float, float], Iterable[Union[float, Tuple[float, float]]]] = None,
        platform: Union[str, Iterable[str]] = None,
        polarization: Union[str, Iterable[str]] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Union[str, Iterable[str]] = None,
        product_list: Union[str, Iterable[str]] = None,
        relativeOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        season: Tuple[int, int] = None,
        start: Union[datetime.datetime, str] = None,
        maxResults: int = None,
        host: str = INTERNAL.SEARCH_API_HOST,
        cmr_token: str = None,
        cmr_provider: str = None
) -> ASFSearchResults:
    """
    Performs a generic search using the ASF SearchAPI

    :param absoluteOrbit: For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B this value corresponds to the orbit count within the orbit cycle. For UAVSAR it is the Flight ID.
    :param asfFrame: This is primarily an ASF / JAXA frame reference. However, some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    :param beamMode: The beam mode used to acquire the data.
    :param campaign: For UAVSAR and AIRSAR data collections only. Search by general location, site description, or data grouping as supplied by flight agency or project.
    :param maxDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param minDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param end: End date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param maxFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param minFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param flightDirection: Satellite orbit direction during data acquisition
    :param flightLine: Specify a flightline for UAVSAR or AIRSAR.
    :param frame: ESA-referenced frames are offered to give users a universal framing convention. Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    :param granule_list: List of specific granules. Search results may include several products per granule name.
    :param groupID: Identifier used to find products considered to be of the same scene but having different granule names
    :param insarStackId: Identifier used to find products of the same InSAR stack
    :param instrument: The instrument used to acquire the data. See also: platform
    :param intersectsWith: Search by polygon, linestring, or point defined in 2D Well-Known Text (WKT)
    :param lookDirection: Left or right look direction during data acquisition
    :param offNadirAngle: Off-nadir angles for ALOS PALSAR
    :param platform: Remote sensing platform that acquired the data. Platforms that work together, such as Sentinel-1A/1B and ERS-1/2 have multi-platform aliases available. See also: instrument
    :param polarization: A property of SAR electromagnetic waves that can be used to extract meaningful information about surface properties of the earth.
    :param processingDate: Used to find data that has been processed at ASF since a given time and date. Supports timestamps as well as natural language such as "3 weeks ago"
    :param processingLevel: Level to which the data has been processed
    :param product_list: List of specific products. Guaranteed to be at most one product per product name.
    :param relativeOrbit: Path or track of satellite during data acquisition. For UAVSAR it is the Line ID.
    :param season: Start and end day of year for desired seasonal range. This option is used in conjunction with start/end to specify a seasonal range within an overall date range.
    :param start: Start date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param maxResults: The maximum number of results to be returned by the search
    :param host: SearchAPI host, defaults to Production SearchAPI. This option is intended for dev/test purposes.
    :param cmr_token: EDL authentication token for authenticated searches, see https://urs.earthdata.nasa.gov/user_tokens
    :param cmr_provider: Custom provider name to constrain CMR results to, for more info on how this is used, see https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#c-provider

    :return: ASFSearchResults(list) of search results
    """
    
    kwargs = locals()
    data = dict((k,v) for k,v in kwargs.items() if v is not None and v != '')
    host = data.pop('host')

    if 'collectionName' in data:
        stack_level = 2
        if inspect.stack()[1].function == 'geo_search':
            stack_level = 3

        warnings.filterwarnings('once')
        warnings.warn("search parameter \"collectionName\" is deprecated and will be removed in a future release. Use \"campaign\" instead.", 
                      DeprecationWarning, 
                      stacklevel=stack_level)
    
    rename_fields = [(
        'campaign', 'collectionName'
    )]
    for (key, replacement) in rename_fields:
        if key in data:
            data[replacement] = data[key]
            data.pop(key)
    
    listify_fields = [
        'absoluteOrbit',
        'asfFrame',
        'beamMode',
        'collectionName',
        'frame',
        'granule_list',
        'groupID',
        'instrument',
        'lookDirection',
        'offNadirAngle',
        'platform',
        'polarization',
        'processingLevel',
        'product_list',
        'relativeOrbit'
    ]
    for key in listify_fields:
        if key in data and not isinstance(data[key], list):
            data[key] = [data[key]]

    flatten_fields = [
        'absoluteOrbit',
        'asfFrame',
        'frame',
        'offNadirAngle',
        'relativeOrbit']
    for key in flatten_fields:
        if key in data:
            data[key] = flatten_list(data[key])

    join_fields = [
        'beamMode',
        'collectionName',
        'flightDirection',
        'granule_list',
        'groupID',
        'instrument',
        'lookDirection',
        'platform',
        'polarization',
        'processingLevel',
        'product_list']
    for key in join_fields:
        if key in data:
            data[key] = ','.join(data[key])

    data['output'] = 'asf_search'

    headers = {'User-Agent': f'{__name__}.{__version__}'}
    response = requests.post(f'https://{host}{INTERNAL.SEARCH_PATH}', data=data, headers=headers)

    try:
        response.raise_for_status()
    except HTTPError:
        if 400 <= response.status_code <= 499:
            raise ASFSearch4xxError(f'HTTP {response.status_code}: {response.json()["error"]["report"]}')
        if 500 <= response.status_code <= 599:
            raise ASFSearch5xxError(f'HTTP {response.status_code}: {response.json()["error"]["report"]}')
        raise ASFServerError(f'HTTP {response.status_code}: {response.json()["error"]["report"]}')

    products = [ASFProduct(f) for f in response.json()['features']]
    return ASFSearchResults(products)


def flatten_list(items: Iterable[Union[float, Tuple[float, float]]]) -> str:
    """
    Converts a list of numbers and/or min/max tuples to a string of comma-separated numbers and/or ranges.
    Example: [1,2,3,(10,20)] -> '1,2,3,10-20'

    :param items: The list of numbers and/or min/max tuples to flatten

    :return: String containing comma-separated representation of input, min/max tuples converted to 'min-max' format

    :raises ValueError: if input list contains tuples with fewer or more than 2 values, or if a min/max tuple in the input list is descending
    :raises TypeError: if input list contains non-numeric values
    """

    for item in items:
        if isinstance(item, tuple):
            if len(item) < 2:
                raise ValueError(f'Not enough values in min/max tuple: {item}')
            if len(item) > 2:
                raise ValueError(f'Too many values in min/max tuple: {item}')
            if not isinstance(item[0], (int, float, complex)) and not isinstance(item[0], bool):
                raise TypeError(f'Expected numeric min in tuple, got {type(item[0])}: {item}')
            if not isinstance(item[1], (int, float, complex)) and not isinstance(item[1], bool):
                raise TypeError(f'Expected numeric max in tuple, got {type(item[1])}: {item}')
            if math.isinf(item[0]) or math.isnan(item[0]):
                raise ValueError(f'Expected finite numeric min in min/max tuple, got {item[0]}: {item}')
            if math.isinf(item[1]) or math.isnan(item[1]):
                raise ValueError(f'Expected finite numeric max in min/max tuple, got {item[1]}: {item}')
            if item[0] > item[1]:
                raise ValueError(f'Min must be less than max when using min/max tuples to search: {item}')
        elif isinstance(item, (int, float, complex)) and not isinstance(item, bool):
            if math.isinf(item) or math.isnan(item):
                raise ValueError(f'Expected finite numeric value, got {item}')
        elif not isinstance(item, (int, float, complex)) and not isinstance(item, bool):
            raise TypeError(f'Expected number or min/max tuple, got {type(item)}')

    return ','.join([f'{item[0]}-{item[1]}' if isinstance(item, tuple) else f'{item}' for item in items])



import requests
import http.cookiejar
from asf_search import __version__
from asf_search.constants import EDL_CLIENT_ID, EDL_HOST, ASF_AUTH_HOST
from asf_search.exceptions import ASFAuthenticationError


class ASFSession(requests.Session):
    AUTH_DOMAINS = ['asf.alaska.edu', 'earthdata.nasa.gov']

    def __init__(self):
        super().__init__()
        self.headers.update({'User-Agent': f'{__name__}.{__version__}'})

    def auth_with_creds(self, username: str, password: str):
        """
        Authenticates the session using EDL username/password credentials

        :param username: EDL username, see https://urs.earthdata.nasa.gov/
        :param password: EDL password, see https://urs.earthdata.nasa.gov/

        :return ASFSession: returns self for convenience
        """
        login_url = f'https://{EDL_HOST}/oauth/authorize?client_id={EDL_CLIENT_ID}&response_type=code&redirect_uri=https://{ASF_AUTH_HOST}/login'

        self.auth = (username, password)
        self.get(login_url)

        if "urs_user_already_logged" not in self.cookies.get_dict():
            raise ASFAuthenticationError("Username or password is incorrect")

        return self

    def auth_with_token(self, token: str):
        """
        Authenticates the session using an EDL Authorization: Bearer token

        :param token: EDL Auth Token for authenticated downloads, see https://urs.earthdata.nasa.gov/user_tokens

        :return ASFSession: returns self for convenience
        """
        self.headers.update({'Authorization': 'Bearer {0}'.format(token)})

        url = "https://cmr.earthdata.nasa.gov/search/collections"
        response = self.get(url)        

        if not 200 <= response.status_code <= 299:
            raise ASFAuthenticationError("Invalid/Expired token passed")

        return self

    def auth_with_cookiejar(self, cookies: http.cookiejar):
        """
        Authenticates the session using a pre-existing cookiejar

        :param cookies: Any http.cookiejar compatible object

        :return ASFSession: returns self for convenience
        """
        
        if "urs_user_already_logged" not in cookies:
            raise ASFAuthenticationError("Cookiejar does not contain login cookies")

        for cookie in cookies:
            if cookie.is_expired():
                raise ASFAuthenticationError("Cookiejar contains expired cookies")

        self.cookies = cookies

        return self

    def rebuild_auth(self, prepared_request: requests.Request, response: requests.Response):
        """
        Overrides requests.Session.rebuild_auth() default behavior of stripping the Authorization header
        upon redirect. This allows token authentication to work with redirects to trusted domains
        """

        headers = prepared_request.headers
        url = prepared_request.url

        if 'Authorization' in headers:
            original_domain = '.'.join(self._get_domain(response.request.url).split('.')[-3:])
            redirect_domain = '.'.join(self._get_domain(url).split('.')[-3:])

            if (original_domain != redirect_domain 
                and (original_domain not in self.AUTH_DOMAINS
                or redirect_domain not in self.AUTH_DOMAINS)):
                del headers['Authorization']

    def _get_domain(self, url: str):
            return requests.utils.urlparse(url).hostname



"""
Simple example script showing a few basic uses of asf_search
"""

import json
import asf_search as asf

print('='*80)
print('Constants')
print(f'asf.BEAMMODE.IW: {asf.BEAMMODE.IW}')
print(f'asf.POLARIZATION.HH_HV: {asf.POLARIZATION.HH_HV}')
print(f'asf.PLATFORM.SENTINEL1: {asf.PLATFORM.SENTINEL1}')

print('='*80)
print(f'Health check: {json.dumps(asf.health(), indent=2)}')

print('='*80)
results = asf.search(platform=[asf.PLATFORM.SENTINEL1], maxResults=2)
print(f'Basic search example: {results}')

print('='*80)
results = asf.granule_search(['ALPSRS279162400', 'ALPSRS279162200'])
print(f'Granule search example: {results}')

print('='*80)
results = asf.product_search(['ALAV2A279102730', 'ALAV2A279133150'])
print(f'Product search example: {results}')

print('='*80)
wkt = 'POLYGON((-135.7 58.2,-136.6 58.1,-135.8 56.9,-134.6 56.1,-134.9 58.0,-135.7 58.2))'
results = asf.geo_search(platform=[asf.PLATFORM.SENTINEL1], intersectsWith=wkt, maxResults=2)
print(f'Geographic search example: {results}')

print('='*80)
results = asf.search(
    platform=[asf.PLATFORM.SENTINEL1],
    frame=[100, 150, (200, 205)],
    relativeOrbit=[100, 105, (110, 115)],
    processingLevel=[asf.PRODUCT_TYPE.SLC])
print(f'Path/frame/platform/product type example: {results}')

print('='*80)
results = asf.stack_from_id('S1B_WV_SLC__1SSV_20210126T234925_20210126T235632_025332_030462_C733-SLC')
print(f'Baseline stack search example, ephemeris-based: {results}')

print('='*80)
try:
    results = asf.stack_from_id('nonexistent-scene')
except asf.ASFSearchError as e:
    print(f'Stacking a non-existent scene throws an exception: {e}')

print('='*80)
try:
    results = asf.stack_from_id('UA_atchaf_06309_21024_020_210401_L090_CX_01-PROJECTED')
except asf.ASFBaselineError as e:
    print(f'Not everything can be stacked: {e}')

print('='*80)
results = asf.stack_from_id('ALPSRP279071390-RTC_HI_RES')
print(f'Baseline stack search example, pre-calculated: {results}')

print('='*80)
results = results[0].stack()
print(f'Baseline stacks can also be made from an ASFProduct: {results}')

print('='*80)
print(f'ASFSearchResults work like lists: {results[3:5]}')

print('='*80)
print(f'ASFSearchResults serializes to geojson: {results[3:5]}')

print('='*80)
product = results[2]
print(f'ASFProduct serializes to geojson: {product}')


print('='*80)
wkt = 'POLYGON((-160 65,-150 65,-160 60,-150 60,-160 65))' # Self-intersecting bowtie
try:
    results = asf.geo_search(platform=[asf.PLATFORM.SENTINEL1], intersectsWith=wkt)
except asf.ASFSearch4xxError as e:
    print(f'Exception example: {e}')

print('='*80)
print('A few more exception examples:')
try:
    asf.search(offNadirAngle=[tuple([1])])
except ValueError as e:
    print(f'Tuple too short: {e}')
try:
    asf.search(offNadirAngle=[(1, 2, 3)])
except ValueError as e:
    print(f'Tuple too long: {e}')
try:
    asf.search(offNadirAngle=[('a', 2)])
except TypeError as e:
    print(f'Tuple non-numeric min: {e}')
try:
    asf.search(offNadirAngle=[(1, 'b')])
except TypeError as e:
    print(f'Tuple non-numeric max: {e}')
try:
    asf.search(offNadirAngle=[(float("NaN"), 2)])
except ValueError as e:
    print(f'Tuple non-finite min: {e}')
try:
    asf.search(offNadirAngle=[1, (float("Inf"))])
except ValueError as e:
    print(f'Tuple non-finite max: {e}')
try:
    asf.search(offNadirAngle=[(2, 1)])
except ValueError as e:
    print(f'Tuple min > max: {e}')
try:
    asf.search(offNadirAngle=[float("Inf")])
except ValueError as e:
    print(f'Bare value non-finite: {e}')
try:
    asf.search(offNadirAngle=['a'])
except TypeError as e:
    print(f'Bare value non-numeric: {e}')



from typing import Iterable
from multiprocessing import Pool
import os.path
import urllib.parse
from requests.exceptions import HTTPError
import warnings

from asf_search.exceptions import ASFAuthenticationError, ASFDownloadError, ASFSearch4xxError
from asf_search import ASFSession


def _download_url(arg):
    url, path, session = arg
    download_url(
        url=url,
        path=path,
        session=session)


def download_urls(urls: Iterable[str], path: str, session: ASFSession = None, processes: int = 1):
    """
    Downloads all products from the specified URLs to the specified location.

    :param urls: List of URLs from which to download
    :param path: Local path in which to save the product
    :param session: The session to use, in most cases should be authenticated beforehand
    :param processes: Number of download processes to use. Defaults to 1 (i.e. sequential download)
    :return:
    """
    if session is None:
        session = ASFSession()

    if processes <= 1:
        for url in urls:
            download_url(url=url, path=path, session=session)
    else:
        pool = Pool(processes=processes)
        args = [(url, path, session) for url in urls]
        pool.map(_download_url, args)
        pool.close()
        pool.join()


def download_url(url: str, path: str, filename: str = None, session: ASFSession = None) -> None:
    """
    Downloads a product from the specified URL to the specified location and (optional) filename.

    :param url: URL from which to download
    :param path: Local path in which to save the product
    :param filename: Optional filename to be used, extracted from the URL by default
    :param session: The session to use, in most cases should be authenticated beforehand
    :return:
    """

    if filename is None:
        filename = os.path.split(urllib.parse.urlparse(url).path)[1]

    if not os.path.isdir(path):
        raise ASFDownloadError(f'Error downloading {url}: directory not found: {path}')

    if os.path.isfile(os.path.join(path, filename)):
        warnings.warn(f'File already exists, skipping download: {os.path.join(path, filename)}')
        return

    if session is None:
        session = ASFSession()


    def strip_auth_if_aws(r, *args, **kwargs):
        if 300 <= r.status_code <= 399 and 'amazonaws.com' in urllib.parse.urlparse(r.headers['location']).netloc:
            location = r.headers['location']
            r.headers.clear()
            r.headers['location'] = location

    response = session.get(url, stream=True, hooks={'response': strip_auth_if_aws})

    try:
        response.raise_for_status()
    except HTTPError as e:
        if 400 <= response.status_code <= 499:
            raise ASFAuthenticationError(f'HTTP {e.response.status_code}: {e.response.text}')
        
        raise e   

    with open(os.path.join(path, filename), 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)



SEARCH_API_HOST = 'api.daac.asf.alaska.edu'
ASF_AUTH_HOST = 'auth.asf.alaska.edu'

HEALTH_PATH = '/health'
SEARCH_PATH = '/services/search/param'
BASELINE_PATH = '/services/search/baseline'
MISSION_PATH = '/services/utils/mission_list'
DATE_PATH = '/services/utils/date'

CMR_HOST = 'cmr.earthdata.nasa.gov'
CMR_COLLECTIONS = '/search/collections.umm_json_v1_4'
EDL_HOST = 'urs.earthdata.nasa.gov'
EDL_CLIENT_ID = 'BO_n7nTIlMljdvU6kRRB3g'



from asf_search.search.search import ASFProduct, ASFSearchResults
from unittest.mock import patch

def run_test_ASFProduct_Geo_Search(geographic_response):
    product = ASFProduct(geographic_response)

    geojson = product.geojson()
    assert(geojson['geometry'] == geographic_response['geometry'])
    assert(geojson['properties'] == geographic_response['properties'])

def run_test_stack(reference, pre_processed_stack, processed_stack):
    product = ASFProduct(reference)
    
    with patch('asf_search.baseline_search.search') as search_mock:
        search_mock.return_value = ASFSearchResults(map(ASFProduct, pre_processed_stack))
        stack = product.stack()


        stack = [
            product for product in stack if product.properties['temporalBaseline'] != None and product.properties['perpendicularBaseline'] != None
            ]

        for(idx, secondary) in enumerate(stack):
            
            if(idx > 0):
                assert(secondary.properties['temporalBaseline'] >= stack[idx - 1].properties['temporalBaseline'])
            
            assert(secondary.properties['temporalBaseline'] == processed_stack[idx]['properties']['temporalBaseline'])
            assert(secondary.properties['perpendicularBaseline'] == processed_stack[idx]['properties']['perpendicularBaseline'])


class ASFError(Exception):
    """Base ASF Exception, not intended for direct use"""

class ASFSearchError(ASFError):
    """Base search-related Exception"""

class ASFSearch4xxError(ASFSearchError):
    """Raise when SearchAPI returns a 4xx error"""

class ASFSearch5xxError(ASFSearchError):
    """Raise when SearchAPI returns a 5xx error"""

class ASFServerError(ASFSearchError):
    """Raise when SearchAPI returns an unknown error"""

class ASFBaselineError(ASFSearchError):
    """Raise when baseline related errors occur"""

class ASFDownloadError(ASFError):
    """Base download-related Exception"""

class ASFAuthenticationError(ASFError):
    """Base download-related Exception"""

class CMRError(Exception):
    """Base CMR Exception"""
